/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package xgi.ut.dsl

import scala.util.parsing.combinator.JavaTokenParsers

/**
 * This contains the grammar of an internal DSL that looks like json
 * but lets us define java object in a simple (short) way.
 */
trait JsonParsers extends JavaTokenParsers {
  import JsonParsers._
  
  /* parses a json structure */
  def jsonstruct:Parser[JsonStructure] = "{"~repsep(jsonprop, ",")~"}"~opt("as"~declaredType) ^^ { 
    case "{"~members~"}"~Some("as"~declType) => JsonStructure(members, Some(declType))
    case "{"~members~"}"~None => JsonStructure(members)
  }
  
  /* parses a json property */
  def jsonprop:Parser[JsonProperty] = jsonid~":"~jsonvalue ^^ { case k~":"~v => JsonProperty(k, v)}
  /* parses a json array */
  def jsonarray:Parser[JsonArray] = "["~>repsep(jsonvalue, ",")<~"]" ^^ {x => JsonArray(x)}
  /* parses a json simple/basic type */
  def jsonbasic:Parser[JsonBasicType] = ( jsontypedstring
                                  | stringLiteral ^^ {x => JsonBasicType(x.drop(1).dropRight(1))} 
                                  | floatingPointNumber ^^ {x =>
                                      val WholeNumber = """(-?\d+)""".r
                                      x match {
                                        case WholeNumber(y) => JsonBasicType(x.toInt)
                                        case other => JsonBasicType(x.toDouble)
                                      }
                                    }
                                  | "true|false".r ^^ {x=>JsonBasicType(x.toBoolean)})
  /* parse an identifier */
  def jsonid:Parser[String] = ident
  
  /* parses a json literal */
  def jsonvalue:Parser[JsonLiteral] = jsonbasic | jsonarray | jsonstruct
  
  /* parses a string that needs to be custom-parsed */
  def jsontypedstring:Parser[JsonBasicType] = stringLiteral~"as"~declaredType ^^ {
    case value~"as"~declaredType => JsonBasicType(value.drop(1).dropRight(1), Some(declaredType))
  }
  
  /* translates a text into an instance of a json literal */
  def parseJson(text:String):JsonLiteral = parseAll(jsonvalue, text) match {
    case Success(x, _) => x
    case NoSuccess(msg, _) => JsonSyntaxError(msg)
  }
  
  def declaredType:Parser[DeclaredType] = stringLiteral ^^ { x => DeclaredType(x.drop(1).dropRight(1))}
}

/* the companion object */
object JsonParsers {
  // an anonymous variable
  sealed case class JsonLiteral
  sealed case class JsonArray(values:Seq[JsonLiteral])            								extends JsonLiteral
  sealed case class JsonProperty(key:String, value:JsonLiteral)  								extends JsonLiteral
  
  // needed to avoid ambiguity in the basic type (string) hierarchy
  sealed case class JsonBasicType(value:Any, declaredType:Option[DeclaredType] = None)                  extends JsonLiteral

  // needed to avoid ambiguity in the json structure hierarchy 
  sealed case class JsonStructure(members:Seq[JsonProperty], declaredType:Option[DeclaredType] = None)  extends JsonLiteral
  
  // mechanism to define whether an object is typed or not
  sealed case class DeclaredType(name:String)
  
  // Not an element from the grammar : nothing but a place holder for the error message
  sealed case class JsonSyntaxError(message:String)                                             extends JsonLiteral
}