/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package xgi.ut.dsl

import scala.annotation.migration
import ScenarParsers.Assignment
import ScenarParsers.Declaration
import ScenarParsers.FieldRef
import ScenarParsers.FnInvocation
import ScenarParsers.InternalIdentifier
import ScenarParsers.MemberRef
import ScenarParsers.SyntaxError
import JsonParsers._
import ConversionConfigParsers._

/**
 * This contains the grammar of an internal DSL meant to simplify our lives 
 * when it comes to configuring unit tests scenarios with a framework such as (ie)
 * Mockito.
 */
trait ScenarParsers extends JsonParsers with ConversionConfigParsers {
  import ScenarParsers._
  /* parses a sequence of instructions separated by a semicolon */
  def scenario:Parser[Seq[Instruction]] = rep1sep(instruction, ";")
  
  /* parses an instruction */
  def instruction:Parser[Instruction] = declaration | assignment | conversionconfinstr
  
  /* parses a conversion config instruction */
  def conversionconfinstr:Parser[ConversionConfigInstruction] = conversionconf ^^ ConversionConfigInstruction
  
  /* parses a value object => internal id or json object */
  def valueObject:Parser[ValueObject] = ( jsonvalue ^^ JsonObject
                                        | memberRef
                                        | mockObject
                                        | internalIdentifier
                                        | fieldRef)
  
  /* declares an internal identifier (aka a variable that does not exists outside the scenario) */
  def declaration:Parser[Declaration] = internalIdentifier~"="~valueObject ^^ { case ref~"="~value => Declaration(ref.name, value)}
  /* parses an assignment (aka the assignment of a value to a given execution chain) */
  def assignment:Parser[Assignment] = memberRef~"="~valueObject ^^ {case ref~"="~value => Assignment(ref, value)} 
  /* parses a reference */
  def reference:Parser[Reference]= fnInvocation | internalIdentifier | fieldRef

  /* parses an internal identifier (aka a variable that does not exists outside the scenario) */
  def internalIdentifier:Parser[InternalIdentifier]="$"~>ident ^^ InternalIdentifier
  
  /* mocks an object */
  def mockObject:Parser[MockObject]= "mock"~>declaredType ^^ MockObject
  
  /* parses a field */
  def fieldRef:Parser[FieldRef]= ident ^^ FieldRef
  /* parses a method call */
  def memberRef:Parser[MemberRef] = (fieldRef | internalIdentifier)~"."~repsep(reference, ".") ^^ {
    case source~"."~invocations => MemberRef(source, invocations)
  }
  /* parses the invocation bit of a method call */
  def fnInvocation:Parser[FnInvocation]=ident~"("~repsep(valueObject, ",")~")" ^^ {
    case fname~"("~args~")" => FnInvocation(fname, args)
  }
  
  /* parses a scenario (sequence of instructions) */
  def parseScenario(script:String):Seq[Instruction] = parseAll(scenario, script) match {
    case Success(scn, _) => scn
    case NoSuccess(err, _) => Seq(SyntaxError(err))
  }
}
/** the companion object */
object ScenarParsers {
  
  sealed case class Instruction
  sealed case class ConversionConfigInstruction(config:ConversionConfig)                          extends Instruction
  sealed case class SyntaxError(message:String)                                                   extends Instruction
  
  sealed case class Declaration(name:String, value:Reference)                                     extends Instruction
  sealed case class Assignment(destination:MemberRef, value:Reference)                            extends Instruction
  
  sealed abstract case class Reference
  
  sealed abstract case class ValueObject														  extends Reference
  sealed case class InternalIdentifier(name:String)                                               extends ValueObject
  sealed case class JsonObject(value:JsonLiteral)												  extends ValueObject
  sealed case class MockObject(declaredType:DeclaredType)                                         extends ValueObject
  
  sealed case class FieldRef(name:String)                                                         extends ValueObject
  sealed case class MemberRef(source:Reference, invocations:Seq[Reference])                       extends ValueObject
  sealed case class FnInvocation(fname:String, params:Seq[Reference])                             extends Reference
}