/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package xgi.ut.dsl.interpreters

import scala.collection.JavaConversions.seqAsJavaList
import scala.collection.mutable.HashMap

import org.mockito.Mockito.mock

import xgi.ut.dsl.JsonParsers._
import xgi.ut.dsl.ScenarParsers._
import xgi.ut.dsl.ScenarParsers.Assignment
import xgi.ut.dsl.ScenarParsers.Declaration
import xgi.ut.dsl.ScenarParsers.FnInvocation
import xgi.ut.dsl.ScenarParsers.Instruction
import xgi.ut.dsl.ScenarParsers.InternalIdentifier
import xgi.ut.dsl.ScenarParsers.JsonObject
import xgi.ut.dsl.ScenarParsers.SyntaxError
import xgi.ut.dsl.utl.ReflectionUtil.invoke

/**
 * This trait provides the capability of interpreting a ScenarLang script.
 */
trait ScenarioInterpreter extends ToObject {
  this:{val context:Any} =>
  val variables = HashMap.empty[String, Any]
  
  /** interprets one instruction */
  def interpret(i:Instruction):Any = i match {
    case SyntaxError(msg) => Console.err.println(msg); throw new RuntimeException(msg)
    case ConversionConfigInstruction(config) => addConfig(config)
    case Declaration(name, value) => {
      val valueObject = dereference(value, context)
      variables += name->valueObject
      valueObject
    }
    case Assignment(destination, value) => {
      val valueObject = dereference(value, context)
      val source = dereference(destination.source, context)
      val interpretation = (source /: destination.invocations.dropRight(1))((newctx, elem) => dereference(elem, newctx))
      assignResult(interpretation, destination.invocations.last, valueObject)
      valueObject
    }
  }
  
  /** dereferences a reference */
  private[this] def dereference(ref:Reference, currentContext:Any):Any = ref match {
    case JsonObject(x) => toObject(x)
    case MockObject(DeclaredType(typeName)) => mock(Class.forName(typeName))
    case InternalIdentifier(name) => variables(name)
    case MemberRef(source, invocations) => {
      val src = dereference(source, currentContext)
      (src /: invocations)((ctx, elem) => dereference(elem, ctx))
    }
    case FieldRef(name) => {
      var externalValue = getFieldValue(currentContext, name)
      // avoid NPE
      if(externalValue==null){
        externalValue = mock(getFieldType(currentContext, name))
        setField(currentContext, name, externalValue)
      }
      externalValue
    }
    case FnInvocation(fname, params) => {
      val args = params map{dereference(_, currentContext)}
      invoke(currentContext, fname, args, null) // is that ok ?
    }
  }
  
  /** assigns a mock result to a given function invocation */
  private[this] def assignResult(target:Any, ref:Reference, result:Any = null):Any = ref match {
    // an internal would make no sense here
    case FieldRef(name) => {
	  setField(target, name, result)
	  getFieldValue(target, name)
    }
    case FnInvocation(fname, params) => {
      val args = params map{dereference(_, target)}
      invoke(target, fname, args, result)
    }
  }
}