/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package xgi.ut.dsl.interpreters

import java.lang.reflect.Field

import scala.collection.immutable.HashMap

import xgi.ut.dsl.JsonParsers.DeclaredType
import xgi.ut.dsl.JsonParsers.JsonArray
import xgi.ut.dsl.JsonParsers.JsonBasicType
import xgi.ut.dsl.JsonParsers.JsonLiteral
import xgi.ut.dsl.JsonParsers.JsonProperty
import xgi.ut.dsl.JsonParsers.JsonStructure
import xgi.ut.dsl.JsonParsers.JsonSyntaxError
import xgi.ut.dsl.utl.ReflectionUtil
/**
 * This trait gives the mixed class/object the ability to let
 * a JsonLiteral enter the structured+typed world of Java & Scala
 */
trait ToObject extends ConversionConfigInterpreter {
  /**
   * Let a JsonLiteral enter the structured+typed world of Java & Scala.
   * A basic type becomes an instance of this basic type.
   * An array becomes a java array
   * A complextype is transformed into a map where each of the fields is a key.
   */
  def toObject(json:JsonLiteral):Any = json match {
    // syntax error
    case JsonSyntaxError(mesg) => Console.err.println(mesg); mesg
    
    // basic types
    case JsonBasicType(value, None) => value
    case JsonBasicType(text,  Some(DeclaredType(typeName))) => convert(text.toString, Class.forName(typeName))
    
    // array
    case JsonArray(elems) => (elems map toObject) toArray
    
    // structure types
    case JsonStructure(props, None) => {
      var theMap = new HashMap[String, Any]()
      props map {case JsonProperty(k, v) => theMap+= k -> toObject(v)}
      theMap
    }
    case JsonStructure(props, Some(DeclaredType(typeName))) => {
      val clazz = Class.forName(typeName)
      val instance = clazz.newInstance()
      props foreach { x => 
        val value = toObject(x.value) match {
          case y:HashMap[String, Any] => {
            val valueInstance = getFieldType(instance, x.key).newInstance()
            fromMap(valueInstance, y)
          }
          case other => other
        }
        setField(instance, x.key, value)
       }
      instance
    } 
    case other => other
  }
  
  /** injects all the properties in the given object */
  def fromMap(target:Any, values:Map[String, Any]):Any = {
    values foreach {y => setField(target, y._1, y._2)}
    target
  }
  
  // WRAPPERS AROUND JAVA REFLECTION API
  /* finds a field of an instance by name */
  def getField(target:Any, name:String):Field = {
    val field = ReflectionUtil.getAllFields(target.getClass).get(name);
    field.setAccessible(true)
    field
  }
  /* returns the type of the given field */
  def getFieldType(target:Any, name:String):Class[_] = getField(target, name).getType
  /* returns the value of the field identified by name for the instance target */
  def getFieldValue(target:Any, name:String):Any = getField(target, name).get(target)
  
  /* 
   * sets the value of the field identified by name for the instance target
   * 
   *  Note: if the value is a string and we know, the field type, this method will
   *  attempt to convert the value to the appropriate type
   */
  def setField(target:Any, name:String, value:Any) = {
    // some kind of type inference
    val valueToSet = value match {
      case text:String => convert(text, getFieldType(target, name));
      case _ => value
    }
    getField(target, name).set(target, valueToSet)
  }
}